@@ -1,5 +1,6 @@
-# conda is distributed under the terms of the BSD 3-clause license.
-# Consult LICENSE.txt or http://opensource.org/licenses/BSD-3-Clause.
+# -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 """OS-agnostic, system-level binary package manager."""
 from __future__ import absolute_import, division, print_function, unicode_literals
 
@@ -16,7 +17,7 @@ __all__ = (
 )
 
 __name__ = "conda"
-__version__ = "4.5.11"
+__version__ = "4.6.1"
 __author__ = "Anaconda, Inc."
 __email__ = "conda@continuum.io"
 __license__ = "BSD-3-Clause"
@@ -46,7 +47,7 @@ class CondaError(Exception):
     def __str__(self):
         try:
             return text_type(self.message % self._kwargs)
-        except:
+        except Exception:
             debug_message = "\n".join((
                 "class: " + self.__class__.__name__,
                 "message:",
@@ -76,7 +77,10 @@ class CondaMultiError(CondaError):
         super(CondaMultiError, self).__init__(None)
 
     def __repr__(self):
-        return '\n'.join(repr(e) for e in self.errors) + '\n'
+        return '\n'.join(text_type(e)
+                         if isinstance(e, EnvironmentError) and not isinstance(e, CondaError)
+                         else repr(e)
+                         for e in self.errors) + '\n'
 
     def __str__(self):
         return '\n'.join(text_type(e) for e in self.errors) + '\n'
