@@ -1,28 +1,21 @@
-# (c) 2012-2013 Continuum Analytics, Inc. / http://continuum.io
-# All Rights Reserved
-#
-# conda is distributed under the terms of the BSD 3-clause license.
-# Consult LICENSE.txt or http://opensource.org/licenses/BSD-3-Clause.
-
+# -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 from collections import OrderedDict
 import json
 from logging import getLogger
 import os
-from os import listdir
 from os.path import exists, expanduser, isfile, join
 import re
 import sys
-from textwrap import dedent
 
 from .common import print_envs_list, stdout_json
 from .. import CONDA_PACKAGE_ROOT, __version__ as conda_version
-from ..base.context import conda_in_private_env, context, sys_rc_path, user_rc_path
+from ..base.context import conda_in_private_env, context, env_name, sys_rc_path, user_rc_path
 from ..common.compat import iteritems, itervalues, on_win, text_type
 from ..common.url import mask_anaconda_token
-from ..core.envs_manager import env_name
-from ..core.subdir_data import SubdirData
 from ..models.channel import all_channel_urls, offline_keep
 from ..models.match_spec import MatchSpec
 from ..utils import human_bytes
@@ -36,7 +29,7 @@ def get_user_site():  # pragma: no cover
         if not on_win:
             if exists(expanduser('~/.local/lib')):
                 python_re = re.compile('python\d\.\d')
-                for path in listdir(expanduser('~/.local/lib/')):
+                for path in os.listdir(expanduser('~/.local/lib/')):
                     if python_re.match(path):
                         site_dirs.append("~/.local/lib/%s" % path)
         else:
@@ -45,7 +38,7 @@ def get_user_site():  # pragma: no cover
             APPDATA = os.environ[str('APPDATA')]
             if exists(join(APPDATA, 'Python')):
                 site_dirs = [join(APPDATA, 'Python', i) for i in
-                             listdir(join(APPDATA, 'PYTHON'))]
+                             os.listdir(join(APPDATA, 'PYTHON'))]
     except (IOError, OSError) as e:
         log.debug('Error accessing user site directory.\n%r', e)
     return site_dirs
@@ -88,7 +81,7 @@ def pretty_package(prec):
 
 
 def print_package_info(packages):
-
+    from ..core.subdir_data import SubdirData
     results = {}
     for package in packages:
         spec = MatchSpec(package)
@@ -101,14 +94,12 @@ def print_package_info(packages):
             for prec in result:
                 pretty_package(prec)
 
+    print("WARNING: 'conda info package_name' is deprecated.\n"
+          "          Use 'conda search package_name --info'.",
+          file=sys.stderr)
+
 
 def get_info_dict(system=False):
-    try:
-        from ..install import linked_data
-        root_pkgs = linked_data(context.root_prefix)
-    except:  # pragma: no cover
-        root_pkgs = {}
-
     try:
         from requests import __version__ as requests_version
         # These environment variables can influence requests' behavior, along with configuration
@@ -126,12 +117,8 @@ def get_info_dict(system=False):
 
     try:
         from conda_env import __version__ as conda_env_version
-    except:  # pragma: no cover
-        try:
-            cenv = [p for p in itervalues(root_pkgs) if p['name'] == 'conda-env']
-            conda_env_version = cenv[0]['version']
-        except:
-            conda_env_version = "not installed"
+    except Exception:  # pragma: no cover
+        conda_env_version = "not installed"
 
     try:
         import conda_build
@@ -189,7 +176,7 @@ def get_info_dict(system=False):
         netrc_file=netrc_file,
     )
     if on_win:
-        from ..common.platform import is_admin_on_windows
+        from ..common.os.windows import is_admin_on_windows
         info_dict['is_windows_admin'] = is_admin_on_windows()
     else:
         info_dict['UID'] = os.geteuid()
@@ -309,7 +296,7 @@ def execute(args, parser):
             print(json.dumps({"channels": context.channels}))
         return 0
 
-    options = 'envs', 'system', 'license'
+    options = 'envs', 'system'
 
     if args.all or context.json:
         for option in options:
@@ -347,17 +334,5 @@ def execute(args, parser):
                 print("%s: %s" % (name, value))
             print()
 
-    if args.license and not context.json:
-        try:
-            from _license import show_info
-            show_info()  # pragma: no cover
-        except ImportError:
-            print(dedent("""
-                WARNING: could not import _license.show_info
-                # try:
-                # $ conda install -n root _license"""))
-        except Exception as e:  # pragma: no cover
-            log.warn('%r', e)
-
     if context.json:
         stdout_json(info_dict)
