@@ -1,19 +1,23 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 from collections import defaultdict
+from datetime import datetime
 
 from .install import calculate_channel_urls
+from .._vendor.boltons.timeutils import UTC
 from ..base.context import context
 from ..cli.common import stdout_json
+from ..common.compat import text_type
 from ..common.io import Spinner
-from ..compat import text_type
 from ..core.envs_manager import query_all_prefixes
 from ..core.subdir_data import SubdirData
-from ..models.records import PackageRecord
 from ..models.match_spec import MatchSpec
+from ..models.records import PackageRecord
 from ..models.version import VersionOrder
-from ..resolve import dashlist
+from ..common.io import dashlist
 from ..utils import human_bytes
 
 
@@ -21,8 +25,6 @@ def execute(args, parser):
     spec = MatchSpec(args.match_spec)
     if spec.get_exact_value('subdir'):
         subdirs = spec.get_exact_value('subdir'),
-    elif args.platform:
-        subdirs = args.platform,
     else:
         subdirs = context.subdirs
 
@@ -66,6 +68,12 @@ def execute(args, parser):
 
         matches = sorted(SubdirData.query_all(spec, channel_urls, subdirs),
                          key=lambda rec: (rec.name, VersionOrder(rec.version), rec.build))
+    if not matches and spec.get_exact_value("name"):
+        flex_spec = MatchSpec(spec, name="*%s*" % spec.name)
+        if not context.json:
+            print("No match found for: %s. Search: %s" % (spec, flex_spec))
+        matches = sorted(SubdirData.query_all(flex_spec, channel_urls, subdirs),
+                         key=lambda rec: (rec.name, VersionOrder(rec.version), rec.build))
 
     if not matches:
         channels_urls = tuple(calculate_channel_urls(
@@ -88,14 +96,14 @@ def execute(args, parser):
             pretty_record(record)
 
     else:
-        builder = ['# %-13s %15s %15s  %-20s' % (
+        builder = ['# %-18s %15s %15s  %-20s' % (
             "Name",
             "Version",
             "Build",
             "Channel",
         )]
         for record in matches:
-            builder.append('%-15s %15s %15s  %-20s' % (
+            builder.append('%-20s %15s %15s  %-20s' % (
                 record.name,
                 record.version,
                 record.build,
@@ -117,16 +125,22 @@ def pretty_record(record):
     push_line("file name", "fn")
     push_line("name", "name")
     push_line("version", "version")
-    push_line("build string", "build")
+    push_line("build", "build")
     push_line("build number", "build_number")
     builder.append("%-12s: %s" % ("size", human_bytes(record.size)))
-    push_line("arch", "arch")
-    push_line("constrains", "constrains")
-    push_line("platform", "platform")
     push_line("license", "license")
     push_line("subdir", "subdir")
     push_line("url", "url")
     push_line("md5", "md5")
-    builder.append("%-12s: %s" % ("dependencies", dashlist(record.depends)))
+    if record.timestamp:
+        date_str = datetime.fromtimestamp(record.timestamp, UTC).strftime('%Y-%m-%d %H:%M:%S %Z')
+        builder.append("%-12s: %s" % ("timestamp", date_str))
+    if record.track_features:
+        builder.append("%-12s: %s" % ("track_features", dashlist(record.track_features)))
+    if record.constrains:
+        builder.append("%-12s: %s" % ("constraints", dashlist(record.constrains)))
+    builder.append(
+        "%-12s: %s" % ("dependencies", dashlist(record.depends) if record.depends else "[]")
+    )
     builder.append('\n')
     print('\n'.join(builder))
