@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 # Try to keep compat small because it's imported by everything
 # What is compat, and what isn't?
 # If a piece of code is "general" and used in multiple modules, it goes here.
@@ -11,6 +13,7 @@ from operator import methodcaller
 import sys
 
 on_win = bool(sys.platform == "win32")
+on_mac = bool(sys.platform == "darwin")
 
 PY2 = sys.version_info[0] == 2
 PY3 = sys.version_info[0] == 3
@@ -60,6 +63,8 @@ elif PY2:  # pragma: py3 no cover
     from itertools import izip as zip, izip_longest as zip_longest
     JSONDecodeError = ValueError
 
+Mapping = Mapping
+Sequence = Sequence
 StringIO = StringIO
 zip = zip
 zip_longest = zip_longest
@@ -144,6 +149,21 @@ def with_metaclass(Type, skip_attrs=set(('__dict__', '__weakref__'))):
     return _clone_with_metaclass
 
 
+def six_with_metaclass(meta, *bases):
+    """Create a base class with a metaclass."""
+    # This requires a bit of explanation: the basic idea is to make a dummy
+    # metaclass for one level of class instantiation that replaces itself with
+    # the actual metaclass.
+    class metaclass(type):
+
+        def __new__(cls, name, this_bases, d):
+            return meta(name, bases, d)
+
+        @classmethod
+        def __prepare__(cls, name, this_bases):
+            return meta.__prepare__(name, bases)
+    return type.__new__(metaclass, str('temporary_class'), (), {})
+
 
 NoneType = type(None)
 primitive_types = tuple(chain(string_types, integer_types, (float, complex, bool, NoneType)))
