@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 """
 A generalized application configuration utility.
 
@@ -22,35 +24,31 @@ from glob import glob
 from itertools import chain
 from logging import getLogger
 from os import environ, stat
-from os.path import basename, join
+from os.path import basename, join, expandvars
 from stat import S_IFDIR, S_IFMT, S_IFREG
 
 from enum import Enum, EnumMeta
 
-from .compat import (isiterable, iteritems, itervalues, odict, primitive_types, string_types,
-                     text_type, with_metaclass)
+from .compat import (binary_type, isiterable, iteritems, itervalues, odict, primitive_types,
+                     string_types, text_type, with_metaclass)
 from .constants import NULL
 from .path import expand
 from .serialize import yaml_load
 from .. import CondaError, CondaMultiError
-from .._vendor.auxlib.collection import AttrDict, first, frozendict, last, make_immutable
+from .._vendor.auxlib.collection import AttrDict, first, last, make_immutable
 from .._vendor.auxlib.exceptions import ThisShouldNeverHappenError
 from .._vendor.auxlib.type_coercion import TypeCoercionError, typify_data_structure
+from .._vendor.frozendict import frozendict
 from .._vendor.boltons.setutils import IndexedSet
+from .._vendor.toolz import concat, concatv, excepts, merge, unique
 
-try:  # pragma: no cover
-    from cytoolz.dicttoolz import merge
-    from cytoolz.functoolz import excepts
-    from cytoolz.itertoolz import concat, concatv, unique
-except ImportError:  # pragma: no cover
-    from .._vendor.toolz.dicttoolz import merge
-    from .._vendor.toolz.functoolz import excepts
-    from .._vendor.toolz.itertoolz import concat, concatv, unique
 try:  # pragma: no cover
     from ruamel_yaml.comments import CommentedSeq, CommentedMap
+    from ruamel_yaml.reader import ReaderError
     from ruamel_yaml.scanner import ScannerError
 except ImportError:  # pragma: no cover
     from ruamel.yaml.comments import CommentedSeq, CommentedMap  # pragma: no cover
+    from ruamel.yaml.reader import ReaderError
     from ruamel.yaml.scanner import ScannerError
 
 log = getLogger(__name__)
@@ -61,26 +59,33 @@ EMPTY_MAP = frozendict()
 def pretty_list(iterable, padding='  '):  # TODO: move elsewhere in conda.common
     if not isiterable(iterable):
         iterable = [iterable]
-    return '\n'.join("%s- %s" % (padding, item) for item in iterable)
+    try:
+        return '\n'.join("%s- %s" % (padding, item) for item in iterable)
+    except TypeError:
+        return pretty_list([iterable], padding)
 
 
 def pretty_map(dictionary, padding='  '):
     return '\n'.join("%s%s: %s" % (padding, key, value) for key, value in iteritems(dictionary))
 
 
-class LoadError(CondaError):
-    def __init__(self, message, filepath, line, column):
-        self.line = line
-        self.filepath = filepath
-        self.column = column
-        msg = "Load Error: in %s on line %s, column %s. %s" % (filepath, line, column, message)
-        super(LoadError, self).__init__(msg)
-
+def expand_environment_variables(unexpanded):
+    if isinstance(unexpanded, string_types) or isinstance(unexpanded, binary_type):
+        return expandvars(unexpanded)
+    else:
+        return unexpanded
 
 class ConfigurationError(CondaError):
     pass
 
 
+class ConfigurationLoadError(ConfigurationError):
+    def __init__(self, path, message_addition='', **kwargs):
+        message = "Unable to load configuration file.\n  path: %(path)s\n"
+        super(ConfigurationLoadError, self).__init__(message + message_addition, path=path,
+                                                     **kwargs)
+
+
 class ValidationError(ConfigurationError):
 
     def __init__(self, parameter_name, parameter_value, source, msg=None, **kwargs):
@@ -332,8 +337,17 @@ class YamlRawParameter(RawParameter):
                 ruamel_yaml = yaml_load(fh)
             except ScannerError as err:
                 mark = err.problem_mark
-                raise LoadError("Invalid YAML", filepath, mark.line, mark.column)
-        return cls.make_raw_parameters(filepath, ruamel_yaml) or EMPTY_MAP
+                raise ConfigurationLoadError(
+                    filepath,
+                    "  reason: invalid yaml at line %(line)s, column %(column)s",
+                    line=mark.line,
+                    column=mark.column
+                )
+            except ReaderError as err:
+                raise ConfigurationLoadError(filepath,
+                                             "  reason: invalid yaml at position %(position)s",
+                                             position=err.position)
+            return cls.make_raw_parameters(filepath, ruamel_yaml) or EMPTY_MAP
 
 
 def load_file_configs(search_path):
@@ -375,12 +389,13 @@ class Parameter(object):
     _type = None
     _element_type = None
 
-    def __init__(self, default, aliases=(), validation=None):
+    def __init__(self, default, aliases=(), validation=None, expandvars=False):
         self._name = None
         self._names = None
         self.default = default
         self.aliases = aliases
         self._validation = validation
+        self._expandvars = expandvars
 
     def _set_name(self, name):
         # this is an explicit method, and not a descriptor/setter
@@ -437,6 +452,19 @@ class Parameter(object):
     def _merge(self, matches):
         raise NotImplementedError()
 
+    def _expand(self, data):
+        if self._expandvars:
+            # This is similar to conda._vendor.auxlib.type_coercion.typify_data_structure
+            # It could be DRY-er but that would break SRP.
+            if isinstance(data, Mapping):
+                return type(data)((k, expand_environment_variables(v)) for k, v in iteritems(data))
+            elif isiterable(data):
+                return type(data)(expand_environment_variables(v) for v in data)
+            else:
+                return expand_environment_variables(data)
+        else:
+            return data
+
     def __get__(self, instance, instance_type):
         # strategy is "extract and merge," which is actually just map and reduce
         # extract matches from each source in SEARCH_PATH
@@ -445,9 +473,12 @@ class Parameter(object):
             return instance._cache_[self.name]
 
         matches, errors = self._get_all_matches(instance)
+        merged = self._merge(matches) if matches else self.default
+        # We need to expand any environment variables before type casting.
+        # Otherwise e.g. `my_bool_var: $BOOL` with BOOL=True would raise a TypeCoercionError.
+        expanded = self._expand(merged)
         try:
-            result = typify_data_structure(self._merge(matches) if matches else self.default,
-                                           self._element_type)
+            result = typify_data_structure(expanded, self._element_type)
         except TypeCoercionError as e:
             errors.append(CustomValidationError(self.name, e.value, "<<merged>>", text_type(e)))
         else:
@@ -507,7 +538,7 @@ class PrimitiveParameter(Parameter):
     python 2 has long and unicode types.
     """
 
-    def __init__(self, default, aliases=(), validation=None, element_type=None):
+    def __init__(self, default, aliases=(), validation=None, element_type=None, expandvars=False):
         """
         Args:
             default (Any):  The parameter's default value.
@@ -521,7 +552,7 @@ class PrimitiveParameter(Parameter):
         """
         self._type = type(default) if element_type is None else element_type
         self._element_type = self._type
-        super(PrimitiveParameter, self).__init__(default, aliases, validation)
+        super(PrimitiveParameter, self).__init__(default, aliases, validation, expandvars)
 
     def _merge(self, matches):
         important_match = first(matches, self._match_key_is_important, default=None)
@@ -546,7 +577,7 @@ class SequenceParameter(Parameter):
     _type = tuple
 
     def __init__(self, element_type, default=(), aliases=(), validation=None,
-                 string_delimiter=','):
+                 string_delimiter=',', expandvars=False):
         """
         Args:
             element_type (type or Iterable[type]): The generic type of each element in
@@ -559,7 +590,7 @@ class SequenceParameter(Parameter):
         """
         self._element_type = element_type
         self.string_delimiter = string_delimiter
-        super(SequenceParameter, self).__init__(default, aliases, validation)
+        super(SequenceParameter, self).__init__(default, aliases, validation, expandvars)
 
     def collect_errors(self, instance, value, source="<<merged>>"):
         errors = super(SequenceParameter, self).collect_errors(instance, value)
@@ -633,9 +664,9 @@ class MapParameter(Parameter):
     """Parameter type for a Configuration class that holds a map (i.e. dict) of python
     primitive values.
     """
-    _type = dict
+    _type = frozendict
 
-    def __init__(self, element_type, default=None, aliases=(), validation=None):
+    def __init__(self, element_type, default=None, aliases=(), validation=None, expandvars=False):
         """
         Args:
             element_type (type or Iterable[type]): The generic type of each element.
@@ -646,7 +677,8 @@ class MapParameter(Parameter):
 
         """
         self._element_type = element_type
-        super(MapParameter, self).__init__(default or dict(), aliases, validation)
+        default = default and frozendict(default) or frozendict()
+        super(MapParameter, self).__init__(default, aliases, validation, expandvars)
 
     def collect_errors(self, instance, value, source="<<merged>>"):
         errors = super(MapParameter, self).collect_errors(instance, value)
@@ -677,8 +709,8 @@ class MapParameter(Parameter):
                                for match, match_value in relevant_matches_and_values)
         # dump all matches in a dict
         # then overwrite with important matches
-        return merge(concatv((v for _, v in relevant_matches_and_values),
-                             reversed(important_maps)))
+        return frozendict(merge(concatv((v for _, v in relevant_matches_and_values),
+                                        reversed(important_maps))))
 
     def repr_raw(self, raw_parameter):
         lines = list()
@@ -851,7 +883,7 @@ class Configuration(object):
         name = parameter.name.lstrip('_')
         aliases = tuple(alias for alias in parameter.aliases if alias != name)
 
-        description = self.get_descriptions()[name]
+        description = self.get_descriptions().get(name, '')
         et = parameter._element_type
         if type(et) == EnumMeta:
             et = [et]
