@@ -1,31 +1,28 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 from functools import reduce
 from logging import getLogger
 import os
-from os.path import abspath, basename, expanduser, expandvars, join, normpath, split, splitext
+from os.path import abspath, basename, expanduser, expandvars, join, normcase, split, splitext
 import re
 import subprocess
 
 from .compat import PY2, ensure_fs_path_encoding, on_win, string_types
 from .. import CondaError
 from .._vendor.auxlib.decorators import memoize
+from .._vendor.toolz import accumulate, concat, take
 
 try:
     # Python 3
     from urllib.parse import unquote, urlsplit
-    from urllib.request import url2pathname
 except ImportError:  # pragma: no cover
     # Python 2
-    from urllib import unquote, url2pathname  # NOQA
+    from urllib import unquote  # NOQA
     from urlparse import urlsplit  # NOQA
 
-try:
-    from cytoolz.itertoolz import accumulate, concat, take
-except ImportError:  # pragma: no cover
-    from .._vendor.toolz.itertoolz import accumulate, concat, take
-
 log = getLogger(__name__)
 
 PATH_MATCH_REGEX = (
@@ -58,7 +55,10 @@ def paths_equal(path1, path2):
         True
 
     """
-    return normpath(abspath(path1)) == normpath(abspath(path2))
+    if on_win:
+        return normcase(abspath(path1)) == normcase(abspath(path2))
+    else:
+        return abspath(path1) == abspath(path2)
 
 
 @memoize
