@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 from errno import EACCES, ENOENT, EPERM
@@ -8,11 +10,11 @@ from os import listdir
 from os.path import basename, dirname, join
 from tarfile import ReadError
 
-from conda._vendor.auxlib.decorators import memoizemethod
-
 from .path_actions import CacheUrlAction, ExtractPackageAction
 from .. import CondaError, CondaMultiError, conda_signal_handler
 from .._vendor.auxlib.collection import first
+from .._vendor.auxlib.decorators import memoizemethod
+from .._vendor.toolz import concat, concatv, groupby
 from ..base.constants import CONDA_TARBALL_EXTENSION, PACKAGE_CACHE_MAGIC_FILE
 from ..base.context import context
 from ..common.compat import (JSONDecodeError, iteritems, itervalues, odict, string_types,
@@ -22,23 +24,17 @@ from ..common.io import ProgressBar, time_recorder
 from ..common.path import expand, url_to_path
 from ..common.signals import signal_handler
 from ..common.url import path_to_url
-from ..exceptions import NotWritableError
+from ..exceptions import NoWritablePkgsDirError, NotWritableError
 from ..gateways.disk.create import (create_package_cache_directory, extract_tarball,
                                     write_as_json_to_file)
 from ..gateways.disk.delete import rm_rf
 from ..gateways.disk.read import (compute_md5sum, isdir, isfile, islink, read_index_json,
                                   read_index_json_from_tarball, read_repodata_json)
 from ..gateways.disk.test import file_path_is_writable
-from ..models.dist import Dist
 from ..models.match_spec import MatchSpec
-from ..models.records import PackageCacheRecord, PackageRecord, PackageRef
+from ..models.records import PackageCacheRecord, PackageRecord
 from ..utils import human_bytes
 
-try:
-    from cytoolz.itertoolz import concat, concatv, groupby
-except ImportError:  # pragma: no cover
-    from .._vendor.toolz.itertoolz import concat, concatv, groupby  # NOQA
-
 log = getLogger(__name__)
 
 
@@ -65,7 +61,7 @@ class PackageCacheData(object):
     def __init__(self, pkgs_dir):
         self.pkgs_dir = pkgs_dir
         self.__package_cache_records = None
-        self.__is_writable = None
+        self.__is_writable = NULL
 
         self._urls_data = UrlsData(pkgs_dir)
 
@@ -98,7 +94,7 @@ class PackageCacheData(object):
         return self
 
     def get(self, package_ref, default=NULL):
-        assert isinstance(package_ref, PackageRef)
+        assert isinstance(package_ref, PackageRecord)
         try:
             return self._package_cache_records[package_ref]
         except KeyError:
@@ -122,7 +118,7 @@ class PackageCacheData(object):
             return (pcrec for pcrec in itervalues(self._package_cache_records)
                     if param.match(pcrec))
         else:
-            assert isinstance(param, PackageRef)
+            assert isinstance(param, PackageRecord)
             return (pcrec for pcrec in itervalues(self._package_cache_records) if pcrec == param)
 
     def iter_records(self):
@@ -133,10 +129,8 @@ class PackageCacheData(object):
         if pkgs_dirs is None:
             pkgs_dirs = context.pkgs_dirs
 
-        return concat(pcache.query(package_ref_or_match_spec) for pcache in concatv(
-            cls.writable_caches(pkgs_dirs),
-            cls.read_only_caches(pkgs_dirs),
-        ))
+        return concat(pcache.query(package_ref_or_match_spec)
+                      for pcache in cls.all_caches_writable_first(pkgs_dirs))
 
     # ##########################################################################################
     # these class methods reach across all package cache directories (usually context.pkgs_dirs)
@@ -144,7 +138,23 @@ class PackageCacheData(object):
 
     @classmethod
     def first_writable(cls, pkgs_dirs=None):
-        return cls.writable_caches(pkgs_dirs)[0]
+        # Calling this method will *create* a package cache directory if one does not already
+        # exist. Any caller should intend to *use* that directory for *writing*, not just reading.
+        if pkgs_dirs is None:
+            pkgs_dirs = context.pkgs_dirs
+        for pkgs_dir in pkgs_dirs:
+            package_cache = cls(pkgs_dir)
+            i_wri = package_cache.is_writable
+            if i_wri is True:
+                return package_cache
+            elif i_wri is None:
+                # means package cache directory doesn't exist, need to try to create it
+                created = create_package_cache_directory(package_cache.pkgs_dir)
+                if created:
+                    package_cache.__is_writable = True
+                    return package_cache
+
+        raise NoWritablePkgsDirError(pkgs_dirs)
 
     @classmethod
     def writable_caches(cls, pkgs_dirs=None):
@@ -152,10 +162,6 @@ class PackageCacheData(object):
             pkgs_dirs = context.pkgs_dirs
         writable_caches = tuple(filter(lambda c: c.is_writable,
                                        (cls(pd) for pd in pkgs_dirs)))
-        if not writable_caches:
-            # TODO: raise NoWritablePackageCacheError()
-            raise CondaError("No writable package cache directories found in\n"
-                             "%s" % text_type(pkgs_dirs))
         return writable_caches
 
     @classmethod
@@ -166,6 +172,16 @@ class PackageCacheData(object):
                                         (cls(pd) for pd in pkgs_dirs)))
         return read_only_caches
 
+    @classmethod
+    def all_caches_writable_first(cls, pkgs_dirs=None):
+        if pkgs_dirs is None:
+            pkgs_dirs = context.pkgs_dirs
+        pc_groups = groupby(
+            lambda pc: pc.is_writable,
+            (cls(pd) for pd in pkgs_dirs)
+        )
+        return tuple(concatv(pc_groups.get(True, ()), pc_groups.get(False, ())))
+
     @classmethod
     def get_all_extracted_entries(cls):
         package_caches = (cls(pd) for pd in context.pkgs_dirs)
@@ -185,13 +201,12 @@ class PackageCacheData(object):
         # if ProgressiveFetchExtract did its job correctly, what we're looking for
         #   should be the matching dist_name in the first writable package cache
         # we'll search all caches for a match, but search writable caches first
-        caches = concatv(cls.writable_caches(), cls.read_only_caches())
         dist_str = package_ref.dist_str().rsplit(':', 1)[-1]
         pc_entry = next((cache._scan_for_dist_no_channel(dist_str)
-                         for cache in caches if cache), None)
+                         for cache in cls.all_caches_writable_first() if cache), None)
         if pc_entry is not None:
             return pc_entry
-        raise CondaError("No package '%s' found in cache directories." % Dist(package_ref))
+        raise CondaError("No package '%s' found in cache directories." % package_ref.dist_str())
 
     @classmethod
     def tarball_file_in_cache(cls, tarball_path, md5sum=None, exclude_caches=()):
@@ -224,24 +239,22 @@ class PackageCacheData(object):
 
     @property
     def is_writable(self):
-        if self.__is_writable is None:
+        # returns None if package cache directory does not exist / has not been created
+        if self.__is_writable is NULL:
             return self._check_writable()
         return self.__is_writable
 
     def _check_writable(self):
-        if isdir(self.pkgs_dir):
+        magic_file = join(self.pkgs_dir, PACKAGE_CACHE_MAGIC_FILE)
+        if isfile(magic_file):
             i_wri = file_path_is_writable(join(self.pkgs_dir, PACKAGE_CACHE_MAGIC_FILE))
+            self.__is_writable = i_wri
+            log.debug("package cache directory '%s' writable: %s", self.pkgs_dir, i_wri)
         else:
             log.trace("package cache directory '%s' does not exist", self.pkgs_dir)
-            i_wri = create_package_cache_directory(self.pkgs_dir)
-        log.debug("package cache directory '%s' writable: %s", self.pkgs_dir, i_wri)
-        self.__is_writable = i_wri
+            self.__is_writable = i_wri = None
         return i_wri
 
-    def _ensure_exists(self):
-        if not isfile(join(self.pkgs_dir, PACKAGE_CACHE_MAGIC_FILE)):
-            create_package_cache_directory(self.pkgs_dir)
-
     @staticmethod
     def _clean_tarball_path_and_get_md5sum(tarball_path, md5sum=None):
         if tarball_path.startswith('file:/'):
@@ -295,7 +308,7 @@ class PackageCacheData(object):
 
             # try reading info/index.json
             try:
-                index_json_record = read_index_json(extracted_package_dir)
+                raw_json_record = read_index_json(extracted_package_dir)
             except (IOError, OSError, JSONDecodeError) as e:
                 # IOError / OSError if info/index.json doesn't exist
                 # JsonDecodeError if info/index.json is partially extracted or corrupted
@@ -327,7 +340,7 @@ class PackageCacheData(object):
                                 rm_rf(package_tarball_full_path)
                                 rm_rf(extracted_package_dir)
                         try:
-                            index_json_record = read_index_json(extracted_package_dir)
+                            raw_json_record = read_index_json(extracted_package_dir)
                         except (IOError, OSError, JSONDecodeError):
                             # At this point, we can assume the package tarball is bad.
                             # Remove everything and move on.
@@ -335,7 +348,7 @@ class PackageCacheData(object):
                             rm_rf(extracted_package_dir)
                             return None
                     else:
-                        index_json_record = read_index_json_from_tarball(package_tarball_full_path)
+                        raw_json_record = read_index_json_from_tarball(package_tarball_full_path)
                 except (EOFError, ReadError) as e:
                     # EOFError: Compressed file ended before the end-of-stream marker was reached
                     # tarfile.ReadError: file could not be opened successfully
@@ -354,7 +367,7 @@ class PackageCacheData(object):
 
             url = self._urls_data.get_url(package_filename)
             package_cache_record = PackageCacheRecord.from_objects(
-                index_json_record,
+                raw_json_record,
                 url=url,
                 md5=md5,
                 package_tarball_full_path=package_tarball_full_path,
@@ -462,7 +475,6 @@ class ProgressiveFetchExtract(object):
         # otherwise, if we find a match in a non-writable cache, we link it to the first writable
         #   cache, and then extract
         first_writable_cache = PackageCacheData.first_writable()
-        first_writable_cache._ensure_exists()
         pcrec_from_writable_cache = next((
             pcrec for pcrec in concat(pcache.query(pref_or_spec)
                                       for pcache in PackageCacheData.writable_caches())
@@ -537,8 +549,8 @@ class ProgressiveFetchExtract(object):
     def __init__(self, link_prefs):
         """
         Args:
-            link_prefs (Tuple[PackageRef]):
-                A sequence of :class:`PackageRef`s to ensure available in a known
+            link_prefs (Tuple[PackageRecord]):
+                A sequence of :class:`PackageRecord`s to ensure available in a known
                 package cache, typically for a follow-on :class:`UnlinkLinkTransaction`.
                 Here, "available" means the package tarball is both downloaded and extracted
                 to a package directory.
@@ -617,9 +629,9 @@ class ProgressiveFetchExtract(object):
         size = getattr(prec_or_spec, 'size', None)
         size_str = size and human_bytes(size) or ''
         if len(desc) > 0:
-            desc = "%-20.20s | " % (desc)
+            desc = "%-20.20s | " % desc
         if len(size_str) > 0:
-            desc += "%-9s | " % (size_str)
+            desc += "%-9s | " % size_str
 
         progress_bar = ProgressBar(desc, not context.verbosity and not context.quiet, context.json)
 
@@ -684,15 +696,3 @@ def rm_fetched(dist):
 def download(url, dst_path, session=None, md5=None, urlstxt=False, retries=3):
     from ..gateways.connection.download import download as gateway_download
     gateway_download(url, dst_path, md5)
-
-
-class package_cache(object):
-
-    def __contains__(self, dist):
-        return bool(PackageCacheData.first_writable().get(Dist(dist).to_package_ref(), None))
-
-    def keys(self):
-        return (Dist(v) for v in itervalues(PackageCacheData.first_writable()))
-
-    def __delitem__(self, dist):
-        PackageCacheData.first_writable().remove(Dist(dist).to_package_ref())
