@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 from abc import ABCMeta, abstractmethod, abstractproperty
@@ -11,8 +13,10 @@ from uuid import uuid4
 from .envs_manager import USER_ENVIRONMENTS_TXT_FILE, register_env, unregister_env
 from .portability import _PaddingError, update_prefix
 from .prefix_data import PrefixData
+from .. import CondaError
 from .._vendor.auxlib.compat import with_metaclass
 from .._vendor.auxlib.ish import dals
+from .._vendor.toolz import concat
 from ..base.constants import CONDA_TARBALL_EXTENSION
 from ..base.context import context
 from ..common.compat import iteritems, on_win, text_type
@@ -23,9 +27,10 @@ from ..common.path import (get_bin_directory_short_path, get_leaf_directories,
 from ..common.url import has_platform, path_to_url, unquote
 from ..exceptions import CondaUpgradeError, CondaVerificationError, PaddingError, SafetyError
 from ..gateways.connection.download import download
-from ..gateways.disk.create import (compile_pyc, copy, create_hard_link_or_copy,
-                                    create_link, create_python_entry_point, extract_tarball,
-                                    make_menu, write_as_json_to_file)
+from ..gateways.disk.create import (compile_multiple_pyc, copy,
+                                    create_hard_link_or_copy, create_link,
+                                    create_python_entry_point, extract_tarball,
+                                    make_menu, mkdir_p, write_as_json_to_file)
 from ..gateways.disk.delete import rm_rf, try_rmdir_all_empty
 from ..gateways.disk.permissions import make_writable
 from ..gateways.disk.read import (compute_md5sum, compute_sha256sum, islink, lexists,
@@ -38,10 +43,6 @@ from ..models.match_spec import MatchSpec
 from ..models.records import (Link, PackageCacheRecord, PackageRecord, PathDataV1, PathsData,
                               PrefixRecord)
 
-try:
-    from cytoolz.itertoolz import concat, concatv
-except ImportError:  # pragma: no cover
-    from .._vendor.toolz.itertoolz import concat, concatv  # NOQA
 
 log = getLogger(__name__)
 
@@ -90,6 +91,44 @@ class PathAction(object):
         return "%s(%s)" % (self.__class__.__name__, ', '.join(args))
 
 
+@with_metaclass(ABCMeta)
+class MultiPathAction(object):
+
+    _verified = False
+
+    @abstractmethod
+    def verify(self):
+        # if verify fails, it should return an exception object rather than raise
+        #  at the end of a verification run, all errors will be raised as a CondaMultiError
+        # after successful verification, the verify method should set self._verified = True
+        raise NotImplementedError()
+
+    @abstractmethod
+    def execute(self):
+        raise NotImplementedError()
+
+    @abstractmethod
+    def reverse(self):
+        raise NotImplementedError()
+
+    @abstractmethod
+    def cleanup(self):
+        raise NotImplementedError()
+
+    @abstractproperty
+    def target_full_paths(self):
+        raise NotImplementedError()
+
+    @property
+    def verified(self):
+        return self._verified
+
+    def __repr__(self):
+        args = ('%s=%r' % (key, value) for key, value in iteritems(vars(self))
+                if key not in REPR_IGNORE_KWARGS)
+        return "%s(%s)" % (self.__class__.__name__, ', '.join(args))
+
+
 @with_metaclass(ABCMeta)
 class PrefixPathAction(PathAction):
 
@@ -137,33 +176,6 @@ class CreateInPrefixPathAction(PrefixPathAction):
         return join(prfx, win_path_ok(shrt_pth)) if prfx and shrt_pth else None
 
 
-# @with_metaclass(ABCMeta)
-# class CreateLeasedPathAction(CreateInPrefixPathAction):
-#     def __init__(self, transaction_context, package_info, source_prefix, source_short_path,
-#                  target_prefix, target_short_path):
-#         super(CreateLeasedPathAction, self).__init__(transaction_context, package_info,
-#                                                      source_prefix, source_short_path,
-#                                                      target_prefix, target_short_path)
-#         self.leased_path_entry = LeasedPathEntry(
-#             _path=target_short_path,
-#             target_path=self.source_full_path,
-#             target_prefix=source_prefix,
-#             leased_path=self.target_full_path,
-#             package_name=package_info.index_json_record.name,
-#             leased_path_type=self.leased_path_type,
-#         )
-#         self._execute_successful = False
-#
-#     def reverse(self):
-#         if self._execute_successful:
-#             log.trace("reversing leased path creation %s", self.target_full_path)
-#             rm_rf(self.target_full_path)
-#
-#     @abstractproperty
-#     def leased_path_type(self):
-#         raise NotImplementedError()
-
-
 class LinkPathAction(CreateInPrefixPathAction):
 
     @classmethod
@@ -188,9 +200,13 @@ class LinkPathAction(CreateInPrefixPathAction):
 
         def make_file_link_action(source_path_data):
             # TODO: this inner function is still kind of a mess
-            noarch = package_info.index_json_record.noarch
+            noarch = package_info.repodata_record.noarch
             if noarch == NoarchType.python:
                 sp_dir = transaction_context['target_site_packages_short_path']
+                if sp_dir is None:
+                    raise CondaError("Unable to determine python site-packages "
+                                     "dir in target_prefix!\nPlease make sure "
+                                     "python is installed in %s" % target_prefix)
                 target_short_path = get_python_noarch_target_path(source_path_data.path, sp_dir)
             elif noarch is None or noarch == NoarchType.generic:
                 target_short_path = source_path_data.path
@@ -258,7 +274,7 @@ class LinkPathAction(CreateInPrefixPathAction):
             The package for %s located at %s
             appears to be corrupted. The path '%s'
             specified in the package manifest cannot be found.
-            """ % (self.package_info.index_json_record.name,
+            """ % (self.package_info.repodata_record.name,
                    self.package_info.extracted_package_dir,
                    self.source_short_path)))
 
@@ -285,29 +301,11 @@ class LinkPathAction(CreateInPrefixPathAction):
             )
 
         elif source_path_data.path_type == PathType.hardlink:
-            try:
-                reported_sha256 = source_path_data.sha256
-            except AttributeError:
-                reported_sha256 = None
-            source_sha256 = compute_sha256sum(self.source_full_path)
-            if reported_sha256 and reported_sha256 != source_sha256:
-                return SafetyError(dals("""
-                The package for %s located at %s
-                appears to be corrupted. The path '%s'
-                has a sha256 mismatch.
-                  reported sha256: %s
-                  actual sha256: %s
-                """ % (self.package_info.index_json_record.name,
-                       self.package_info.extracted_package_dir,
-                       self.source_short_path,
-                       reported_sha256,
-                       source_sha256,
-                       )))
-
             try:
                 reported_size_in_bytes = source_path_data.size_in_bytes
             except AttributeError:
                 reported_size_in_bytes = None
+            source_size_in_bytes = 0
             if reported_size_in_bytes:
                 source_size_in_bytes = getsize(self.source_full_path)
                 if reported_size_in_bytes != source_size_in_bytes:
@@ -317,13 +315,35 @@ class LinkPathAction(CreateInPrefixPathAction):
                     has an incorrect size.
                       reported size: %s bytes
                       actual size: %s bytes
-                    """ % (self.package_info.index_json_record.name,
+                    """ % (self.package_info.repodata_record.name,
                            self.package_info.extracted_package_dir,
                            self.source_short_path,
                            reported_size_in_bytes,
                            source_size_in_bytes,
                            )))
 
+            try:
+                reported_sha256 = source_path_data.sha256
+            except AttributeError:
+                reported_sha256 = None
+            # sha256 is expensive.  Only run if file sizes agree, and then only if enabled
+            if (source_size_in_bytes and reported_size_in_bytes == source_size_in_bytes
+                    and context.extra_safety_checks):
+                source_sha256 = compute_sha256sum(self.source_full_path)
+
+                if reported_sha256 and reported_sha256 != source_sha256:
+                    return SafetyError(dals("""
+                    The package for %s located at %s
+                    appears to be corrupted. The path '%s'
+                    has a sha256 mismatch.
+                    reported sha256: %s
+                    actual sha256: %s
+                    """ % (self.package_info.repodata_record.name,
+                           self.package_info.extracted_package_dir,
+                           self.source_short_path,
+                           reported_sha256,
+                           source_sha256,
+                           )))
             self.prefix_path_data = PathDataV1.from_objects(
                 source_path_data,
                 sha256=reported_sha256,
@@ -380,6 +400,7 @@ class PrefixReplaceLinkAction(LinkPathAction):
             # return
             assert False, "I don't think this is the right place to ignore this"
 
+        mkdir_p(self.transaction_context['temp_dir'])
         self.intermediate_path = join(self.transaction_context['temp_dir'], text_type(uuid4()))
 
         log.trace("copying %s => %s", self.source_full_path, self.intermediate_path)
@@ -468,7 +489,7 @@ class CreateNonadminAction(CreateInPrefixPathAction):
             rm_rf(self.target_full_path)
 
 
-class CompilePycAction(CreateInPrefixPathAction):
+class CompileMultiPycAction(MultiPathAction):
 
     @classmethod
     def create_actions(cls, transaction_context, package_info, target_prefix, requested_link_type,
@@ -477,41 +498,69 @@ class CompilePycAction(CreateInPrefixPathAction):
         if noarch is not None and noarch.type == NoarchType.python:
             noarch_py_file_re = re.compile(r'^site-packages[/\\][^\t\n\r\f\v]+\.py$')
             py_ver = transaction_context['target_python_version']
-            py_files = (axn.target_short_path for axn in file_link_actions
-                        if noarch_py_file_re.match(axn.source_short_path))
-            return tuple(cls(transaction_context, package_info, target_prefix,
-                             pf, pyc_path(pf, py_ver))
-                         for pf in py_files)
+            py_files = tuple((axn.target_short_path for axn in file_link_actions
+                              if noarch_py_file_re.match(axn.source_short_path)))
+            pyc_files = tuple((pyc_path(pf, py_ver) for pf in py_files))
+            return (cls(transaction_context, package_info, target_prefix, py_files, pyc_files), )
         else:
             return ()
 
     def __init__(self, transaction_context, package_info, target_prefix,
-                 source_short_path, target_short_path):
-        super(CompilePycAction, self).__init__(transaction_context, package_info,
-                                               target_prefix, source_short_path,
-                                               target_prefix, target_short_path)
-        self.prefix_path_data = PathDataV1(
-            _path=self.target_short_path,
-            path_type=PathType.pyc_file,
-        )
+                 source_short_paths, target_short_paths):
+        self.transaction_context = transaction_context
+        self.package_info = package_info
+        self.target_prefix = target_prefix
+        self.source_short_paths = source_short_paths
+        self.target_short_paths = target_short_paths
+        self.prefix_path_data = None
+        self.prefix_paths_data = [
+            PathDataV1(_path=p, path_type=PathType.pyc_file,) for p in self.target_short_paths]
         self._execute_successful = False
 
+    @property
+    def target_full_paths(self):
+        def join_or_none(prefix, short_path):
+            if prefix is None or short_path is None:
+                return None
+            else:
+                return join(prefix, win_path_ok(short_path))
+        return (join_or_none(self.target_prefix, p) for p in self.target_short_paths)
+
+    @property
+    def source_full_paths(self):
+        def join_or_none(prefix, short_path):
+            if prefix is None or short_path is None:
+                return None
+            else:
+                return join(prefix, win_path_ok(short_path))
+        return (join_or_none(self.target_prefix, p) for p in self.source_short_paths)
+
+    def verify(self):
+        self._verified = True
+
+    def cleanup(self):
+        # create actions typically won't need cleanup
+        pass
+
     def execute(self):
         # compile_pyc is sometimes expected to fail, for example a python 3.6 file
         #   installed into a python 2 environment, but no code paths actually importing it
         # technically then, this file should be removed from the manifest in conda-meta, but
         #   at the time of this writing that's not currently happening
-        log.trace("compiling %s", self.target_full_path)
+        log.trace("compiling %s", ' '.join(self.target_full_paths))
         target_python_version = self.transaction_context['target_python_version']
         python_short_path = get_python_short_path(target_python_version)
         python_full_path = join(self.target_prefix, win_path_ok(python_short_path))
-        compile_pyc(python_full_path, self.source_full_path, self.target_full_path)
+        compile_multiple_pyc(python_full_path, self.source_full_paths, self.target_full_paths,
+                             self.target_prefix, self.transaction_context['target_python_version'])
         self._execute_successful = True
 
     def reverse(self):
+        # this removes all pyc files even if they were not created
         if self._execute_successful:
-            log.trace("reversing pyc creation %s", self.target_full_path)
-            rm_rf(self.target_full_path)
+            log.trace("reversing pyc creation %s", ' '.join(self.target_full_paths))
+            for target_full_path in self.target_full_paths:
+                rm_rf(target_full_path)
 
 
 class CreatePythonEntryPointAction(CreateInPrefixPathAction):
@@ -767,6 +816,7 @@ class CreatePrefixRecordAction(CreateInPrefixPathAction):
         self.requested_link_type = requested_link_type
         self.requested_spec = requested_spec
         self.all_link_path_actions = all_link_path_actions
+        self._execute_successful = False
 
     def execute(self):
         link = Link(
@@ -777,17 +827,30 @@ class CreatePrefixRecordAction(CreateInPrefixPathAction):
         package_tarball_full_path = extracted_package_dir + CONDA_TARBALL_EXTENSION
         # TODO: don't make above assumption; put package_tarball_full_path in package_info
 
-        files = (x.target_short_path for x in self.all_link_path_actions if x)
+        def files_from_action(link_path_action):
+            if isinstance(link_path_action, CompileMultiPycAction):
+                return link_path_action.target_short_paths
+            else:
+                return (link_path_action.target_short_path, )
 
+        def paths_from_action(link_path_action):
+            if isinstance(link_path_action, CompileMultiPycAction):
+                return link_path_action.prefix_paths_data
+            else:
+                if link_path_action.prefix_path_data is None:
+                    return ()
+                else:
+                    return (link_path_action.prefix_path_data, )
+
+        files = concat((files_from_action(x) for x in self.all_link_path_actions if x))
         paths_data = PathsData(
             paths_version=1,
-            paths=(x.prefix_path_data for x in self.all_link_path_actions
-                   if x and x.prefix_path_data),
+            paths=concat((paths_from_action(x) for x in self.all_link_path_actions if x)),
         )
 
         self.prefix_record = PrefixRecord.from_objects(
             self.package_info.repodata_record,
-            self.package_info.index_json_record,
+            # self.package_info.index_json_record,
             self.package_info.package_metadata,
             requested_spec=text_type(self.requested_spec),
             paths_data=paths_data,
@@ -800,11 +863,12 @@ class CreatePrefixRecordAction(CreateInPrefixPathAction):
 
         log.trace("creating linked package record %s", self.target_full_path)
         PrefixData(self.target_prefix).insert(self.prefix_record)
+        self._execute_successful = True
 
     def reverse(self):
         log.trace("reversing linked package record creation %s", self.target_full_path)
-        # TODO: be careful about failure here, and being too strict
-        PrefixData(self.target_prefix).remove(self.package_info.index_json_record.name)
+        if self._execute_successful:
+            PrefixData(self.target_prefix).remove(self.package_info.repodata_record.name)
 
 
 class UpdateHistoryAction(CreateInPrefixPathAction):
@@ -872,59 +936,6 @@ class RegisterEnvironmentLocationAction(PathAction):
         raise NotImplementedError()
 
 
-# class RegisterPrivateEnvAction(EnvsDirectoryPathAction):
-#
-#     @classmethod
-#     def create_actions(cls, transaction_context, package_info, target_prefix, requested_spec,
-#                        leased_paths):
-#         preferred_env = package_info.repodata_record.preferred_env
-#         if preferred_env_matches_prefix(preferred_env, target_prefix, context.root_prefix):
-#             return cls(transaction_context, package_info, context.root_prefix, preferred_env,
-#                        requested_spec, leased_paths),
-#         else:
-#             return ()
-#
-#     def __init__(self, transaction_context, package_info, root_prefix, env_name, requested_spec,
-#                  leased_paths):
-#         self.root_prefix = root_prefix
-#         self.env_name = ensure_pad(env_name)
-#         target_prefix = join(self.root_prefix, 'envs', self.env_name)
-#         super(RegisterPrivateEnvAction, self).__init__(transaction_context, target_prefix)
-#
-#         self.package_name = package_info.index_json_record.name
-#         self.requested_spec = requested_spec
-#         self.leased_paths = leased_paths
-#
-#         fn = basename(package_info.extracted_package_dir) + '.json'
-#         self.conda_meta_path = join(self.target_prefix, 'conda-meta', fn)
-#
-#     def execute(self):
-#         log.trace("registering private env for %s", self.target_prefix)
-#
-#         # touches env prefix entry in catalog.json
-#         # updates leased_paths
-#         from .envs_manager import EnvsDirectory
-#         ed = EnvsDirectory(self.envs_dir_path)
-#
-#         self.envs_dir_state = ed._get_state()
-#
-#         for leased_path_entry in self.leased_paths:
-#             ed.add_leased_path(leased_path_entry)
-#
-#         ed.add_preferred_env_package(self.env_name, self.package_name, self.conda_meta_path,
-#                                      self.requested_spec)
-#         ed.write_to_disk()
-#         self._execute_successful = True
-#
-#     def reverse(self):
-#         if self._execute_successful:
-#             log.trace("reversing environment unregistration in catalog for %s", self.target_prefix)  # NOQA
-#             from .envs_manager import EnvsDirectory
-#             ed = EnvsDirectory(self.envs_dir_path)
-#             ed._set_state(self.envs_dir_state)
-#             ed.write_to_disk()
-
-
 # ######################################################
 #  Removal of Paths within a Prefix
 # ######################################################
@@ -1042,62 +1053,6 @@ class UnregisterEnvironmentLocationAction(PathAction):
         raise NotImplementedError()
 
 
-# class UnregisterPrivateEnvAction(EnvsDirectoryPathAction):
-#
-#     @classmethod
-#     def create_actions(cls, transaction_context, linked_package_data, target_prefix):
-#         preferred_env = ensure_pad(linked_package_data.preferred_env)
-#         if preferred_env_matches_prefix(preferred_env, target_prefix, context.root_prefix):
-#             package_name = linked_package_data.name
-#
-#             from .envs_manager import EnvsDirectory
-#             envs_directory_path = EnvsDirectory.get_envs_directory_for_prefix(target_prefix)
-#             ed = EnvsDirectory(envs_directory_path)
-#
-#             ed.get_leased_path_entries_for_package(package_name)
-#
-#             leased_path_entries = ed.get_leased_path_entries_for_package(package_name)
-#             leased_paths_to_remove = tuple(lpe._path for lpe in leased_path_entries)
-#             unlink_leased_path_actions = (UnlinkPathAction(transaction_context, None,
-#                                                            context.root_prefix, lp)
-#                                           for lp in leased_paths_to_remove)
-#
-#             unregister_private_env_actions = cls(transaction_context, context.root_prefix,
-#                                                  package_name),
-#
-#             return concatv(unlink_leased_path_actions, unregister_private_env_actions)
-#
-#         else:
-#             return ()
-#
-#     def __init__(self, transaction_context, root_prefix, package_name):
-#         super(UnregisterPrivateEnvAction, self).__init__(transaction_context, root_prefix)
-#         self.root_prefix = root_prefix
-#         self.package_name = package_name
-#
-#     def execute(self):
-#         log.trace("unregistering private env for %s", self.package_name)
-#
-#         from .envs_manager import EnvsDirectory
-#         ed = EnvsDirectory(self.envs_dir_path)
-#
-#         self.envs_dir_state = ed._get_state()
-#
-#         ed.remove_preferred_env_package(self.package_name)
-#
-#         ed.write_to_disk()
-#         self._execute_successful = True
-#
-#     def reverse(self):
-#         if self._execute_successful:
-#             log.trace("reversing environment unregistration in catalog for %s",
-#                       self.target_prefix)
-#             from .envs_manager import EnvsDirectory
-#             ed = EnvsDirectory(self.envs_dir_path)
-#             ed._set_state(self.envs_dir_state)
-#             ed.write_to_disk()
-
-
 # ######################################################
 #  Fetch / Extract Actions
 # ######################################################
@@ -1241,7 +1196,7 @@ class ExtractPackageAction(PathAction):
         extract_tarball(self.source_full_path, self.target_full_path,
                         progress_update_callback=progress_update_callback)
 
-        index_json_record = read_index_json(self.target_full_path)
+        raw_index_json = read_index_json(self.target_full_path)
 
         if isinstance(self.record_or_spec, MatchSpec):
             url = self.record_or_spec.get_raw_value('url')
@@ -1249,10 +1204,10 @@ class ExtractPackageAction(PathAction):
             channel = Channel(url) if has_platform(url, context.known_subdirs) else Channel(None)
             fn = basename(url)
             md5 = self.md5sum or compute_md5sum(self.source_full_path)
-            repodata_record = PackageRecord.from_objects(index_json_record, url=url,
+            repodata_record = PackageRecord.from_objects(raw_index_json, url=url,
                                                          channel=channel, fn=fn, md5=md5)
         else:
-            repodata_record = PackageRecord.from_objects(self.record_or_spec, index_json_record)
+            repodata_record = PackageRecord.from_objects(self.record_or_spec, raw_index_json)
 
         repodata_record_path = join(self.target_full_path, 'info', 'repodata_record.json')
         write_as_json_to_file(repodata_record_path, repodata_record)
