@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 import bz2
@@ -12,13 +14,14 @@ from logging import DEBUG, getLogger
 from mmap import ACCESS_READ, mmap
 from os.path import dirname, isdir, join, splitext
 import re
-from textwrap import dedent
 from time import time
 import warnings
+from io import open as io_open
 
 from .. import CondaError
 from .._vendor.auxlib.ish import dals
 from .._vendor.auxlib.logz import stringify
+from .._vendor.toolz import concat, take
 from ..base.constants import CONDA_HOMEPAGE_URL
 from ..base.context import context
 from ..common.compat import (ensure_binary, ensure_text_type, ensure_unicode, iteritems,
@@ -26,7 +29,8 @@ from ..common.compat import (ensure_binary, ensure_text_type, ensure_unicode, it
 from ..common.io import ThreadLimitedThreadPoolExecutor, as_completed
 from ..common.url import join_url, maybe_unquote
 from ..core.package_cache_data import PackageCacheData
-from ..exceptions import CondaDependencyError, CondaHTTPError, CondaUpgradeError, NotWritableError
+from ..exceptions import (CondaDependencyError, CondaHTTPError, CondaUpgradeError,
+                          NotWritableError, UnavailableInvalidChannel)
 from ..gateways.connection import (ConnectionError, HTTPError, InsecureRequestWarning,
                                    InvalidSchema, SSLError)
 from ..gateways.connection.session import CondaSession
@@ -34,14 +38,8 @@ from ..gateways.disk import mkdir_p, mkdir_p_sudo_safe
 from ..gateways.disk.delete import rm_rf
 from ..gateways.disk.update import touch
 from ..models.channel import Channel, all_channel_urls
-from ..models.dist import Dist
 from ..models.match_spec import MatchSpec
-from ..models.records import PackageRecord, PackageRef
-
-try:
-    from cytoolz.itertoolz import concat, take
-except ImportError:  # pragma: no cover
-    from .._vendor.toolz.itertoolz import concat, take  # NOQA
+from ..models.records import PackageRecord
 
 try:
     import cPickle as pickle
@@ -51,7 +49,7 @@ except ImportError:  # pragma: no cover
 log = getLogger(__name__)
 stderrlog = getLogger('conda.stderrlog')
 
-REPODATA_PICKLE_VERSION = 19
+REPODATA_PICKLE_VERSION = 28
 MAX_REPODATA_VERSION = 1
 REPODATA_HEADER_RE = b'"(_etag|_mod|_cache_control)":[ ]?"(.*?[^\\\\])"[,\}\s]'
 
@@ -114,7 +112,7 @@ class SubdirData(object):
                     if param.match(prec):
                         yield prec
         else:
-            assert isinstance(param, PackageRef)
+            assert isinstance(param, PackageRecord)
             for prec in self._names_index[param.name]:
                 if prec == param:
                     yield prec
@@ -147,7 +145,7 @@ class SubdirData(object):
 
     def load(self):
         _internal_state = self._load()
-        if _internal_state["repodata_version"] > MAX_REPODATA_VERSION:
+        if _internal_state.get("repodata_version", 0) > MAX_REPODATA_VERSION:
             raise CondaUpgradeError(dals("""
                 The current version of conda is too old to read repodata from
 
@@ -159,7 +157,6 @@ class SubdirData(object):
 
         self._internal_state = _internal_state
         self._package_records = _internal_state['_package_records']
-        self._package_dists = _internal_state['_package_dists']  # only needed as an optimization for conda-build  # NOQA
         self._names_index = _internal_state['_names_index']
         self._track_features_index = _internal_state['_track_features_index']
         self._loaded = True
@@ -170,11 +167,6 @@ class SubdirData(object):
             self.load()
         return iter(self._package_records)
 
-    def iter_dists_records(self):
-        if not self._loaded:
-            self.load()
-        return zip(self._package_dists, self._package_records)
-
     def _load(self):
         try:
             mtime = getmtime(self.cache_path_json)
@@ -186,7 +178,6 @@ class SubdirData(object):
                           self.url_w_subdir, self.cache_path_json)
                 return {
                     '_package_records': (),
-                    '_package_dists': (),
                     '_names_index': defaultdict(list),
                     '_track_features_index': defaultdict(list),
                 }
@@ -236,7 +227,7 @@ class SubdirData(object):
             if not isdir(dirname(self.cache_path_json)):
                 mkdir_p(dirname(self.cache_path_json))
             try:
-                with open(self.cache_path_json, 'w') as fh:
+                with io_open(self.cache_path_json, 'w') as fh:
                     fh.write(raw_repodata_str or '{}')
             except (IOError, OSError) as e:
                 if e.errno in (EACCES, EPERM):
@@ -322,7 +313,6 @@ class SubdirData(object):
         schannel = self.channel.canonical_name
 
         self._package_records = _package_records = []
-        self._package_dists = _package_dists = []  # creating and caching these here is an optimization for conda-build  # NOQA
         self._names_index = _names_index = defaultdict(list)
         self._track_features_index = _track_features_index = defaultdict(list)
 
@@ -333,7 +323,6 @@ class SubdirData(object):
             'cache_path_base': self.cache_path_base,
 
             '_package_records': _package_records,
-            '_package_dists': _package_dists,
             '_names_index': _names_index,
             '_track_features_index': _track_features_index,
 
@@ -371,10 +360,13 @@ class SubdirData(object):
             if add_pip and info['name'] == 'python' and info['version'].startswith(('2.', '3.')):
                 info['depends'].append('pip')
             info.update(meta_in_common)
+            if info.get('record_version', 0) > 1:
+                log.debug("Ignoring record_version %d from %s",
+                          info["record_version"], info['url'])
+                continue
             package_record = PackageRecord(**info)
 
             _package_records.append(package_record)
-            _package_dists.append(Dist(package_record))
             _names_index[package_record.name].append(package_record)
             for ftr_name in package_record.track_features:
                 _track_features_index[ftr_name].append(package_record)
@@ -436,7 +428,7 @@ def fetch_repodata_remote_request(url, etag, mod_stamp):
         resp = session.get(join_url(url, filename), headers=headers, proxies=session.proxies,
                            timeout=timeout)
         if log.isEnabledFor(DEBUG):
-            log.debug(stringify(resp))
+            log.debug(stringify(resp, content_max_len=256))
         resp.raise_for_status()
 
     except InvalidSchema as e:
@@ -456,43 +448,15 @@ def fetch_repodata_remote_request(url, etag, mod_stamp):
         status_code = getattr(e.response, 'status_code', None)
         if status_code in (403, 404):
             if not url.endswith('/noarch'):
-                return None
-            else:
                 if context.allow_non_channel_urls:
-                    help_message = dedent("""
-                    WARNING: The remote server could not find the noarch directory for the
-                    requested channel with url: %s
-
-                    It is possible you have given conda an invalid channel. Please double-check
-                    your conda configuration using `conda config --show channels`.
-
-                    If the requested url is in fact a valid conda channel, please request that the
-                    channel administrator create `noarch/repodata.json` and associated
-                    `noarch/repodata.json.bz2` files, even if `noarch/repodata.json` is empty.
-                    $ mkdir noarch
-                    $ echo '{}' > noarch/repodata.json
-                    $ bzip2 -k noarch/repodata.json
-                    """) % maybe_unquote(dirname(url))
-                    stderrlog.warn(help_message)
+                    stderrlog.warning("Unable to retrieve repodata (%d error) for %s",
+                                      status_code, url)
                     return None
                 else:
-                    help_message = dals("""
-                    The remote server could not find the noarch directory for the
-                    requested channel with url: %s
-
-                    As of conda 4.3, a valid channel must contain a `noarch/repodata.json` and
-                    associated `noarch/repodata.json.bz2` file, even if `noarch/repodata.json` is
-                    empty. please request that the channel administrator create
-                    `noarch/repodata.json` and associated `noarch/repodata.json.bz2` files.
-                    $ mkdir noarch
-                    $ echo '{}' > noarch/repodata.json
-                    $ bzip2 -k noarch/repodata.json
-
-                    You will need to adjust your conda configuration to proceed.
-                    Use `conda config --show channels` to view your configuration's current state.
-                    Further configuration help can be found at <%s>.
-                    """) % (maybe_unquote(dirname(url)),
-                            join_url(CONDA_HOMEPAGE_URL, 'docs/config.html'))
+                    raise UnavailableInvalidChannel(Channel(dirname(url)), status_code)
+            else:
+                log.info("Unable to retrieve repodata (%d error) for %s", status_code, url)
+                return None
 
         elif status_code == 401:
             channel = Channel(url)
@@ -587,12 +551,12 @@ def fetch_repodata_remote_request(url, etag, mod_stamp):
 
     # add extra values to the raw repodata json
     if json_str and json_str != "{}":
-        raw_repodata_str = "%s, %s" % (
+        raw_repodata_str = u"%s, %s" % (
             json.dumps(saved_fields)[:-1],  # remove trailing '}'
             json_str[1:]  # remove first '{'
         )
     else:
-        raw_repodata_str = json.dumps(saved_fields)
+        raw_repodata_str = ensure_text_type(json.dumps(saved_fields))
     return raw_repodata_str
 
 
@@ -612,14 +576,6 @@ def make_feature_record(feature_name):
     )
 
 
-def collect_all_repodata_as_index(use_cache, channel_urls):
-    index = {}
-    for url in channel_urls:
-        sd = SubdirData(Channel(url))
-        index.update(sd.iter_dists_records())
-    return index
-
-
 def cache_fn_url(url):
     # url must be right-padded with '/' to not invalidate any existing caches
     if not url.endswith('/'):
