@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 from datetime import timedelta
@@ -6,23 +8,20 @@ from errno import ENOSPC
 import json
 from logging import getLogger
 import os
+from os.path import join
 import sys
+from textwrap import dedent
 from traceback import format_exception, format_exception_only
 
 from . import CondaError, CondaExitZero, CondaMultiError, text_type
 from ._vendor.auxlib.entity import EntityEncoder
 from ._vendor.auxlib.ish import dals
 from ._vendor.auxlib.type_coercion import boolify
-from .base.constants import PathConflict, SafetyChecks
-from .common.compat import ensure_text_type, input, iteritems, iterkeys, on_win, string_types, PY2
-from .common.io import timeout
+from ._vendor.toolz import groupby
+from .base.constants import COMPATIBLE_SHELLS, PathConflict, SafetyChecks
+from .common.compat import PY2, ensure_text_type, input, iteritems, iterkeys, on_win, string_types
+from .common.io import dashlist, timeout
 from .common.signals import get_signal_name
-from .common.url import maybe_unquote
-
-try:
-    from cytoolz.itertoolz import groupby
-except ImportError:  # pragma: no cover
-    from ._vendor.toolz.itertoolz import groupby  # NOQA
 
 log = getLogger(__name__)
 
@@ -32,8 +31,11 @@ log = getLogger(__name__)
 class ResolvePackageNotFound(CondaError):
     def __init__(self, bad_deps):
         # bad_deps is a list of lists
+        # bad_deps should really be named 'invalid_chains'
         self.bad_deps = tuple(dep for deps in bad_deps for dep in deps if dep)
-        message = '\n' + '\n'.join(('  - %s' % dep) for dep in self.bad_deps)
+        formatted_chains = tuple(" -> ".join(map(str, bad_chain)) for bad_chain in bad_deps)
+        self._formatted_chains = formatted_chains
+        message = '\n' + '\n'.join(('  - %s' % bad_chain) for bad_chain in formatted_chains)
         super(ResolvePackageNotFound, self).__init__(message)
 NoPackagesFound = NoPackagesFoundError = ResolvePackageNotFound  # NOQA
 
@@ -45,16 +47,73 @@ class LockError(CondaError):
 
 
 class ArgumentError(CondaError):
+    return_code = 2
+
     def __init__(self, message, **kwargs):
         super(ArgumentError, self).__init__(message, **kwargs)
 
 
 class CommandArgumentError(ArgumentError):
+    # TODO: Consolidate with ArgumentError.
+    return_code = 2
+
     def __init__(self, message, **kwargs):
         command = ' '.join(ensure_text_type(s) for s in sys.argv)
         super(CommandArgumentError, self).__init__(message, command=command, **kwargs)
 
 
+class Help(CondaError):
+    pass
+
+
+class ActivateHelp(Help):
+
+    def __init__(self):
+        message = dals("""
+        usage: conda activate [-h] [--stack] [env_name_or_prefix]
+
+        Activate a conda environment.
+
+        Options:
+
+        positional arguments:
+          env_name_or_prefix    The environment name or prefix to activate. If the
+                                prefix is a relative path, it must start with './'
+                                (or '.\\' on Windows).
+
+        optional arguments:
+          -h, --help            Show this help message and exit.
+          --stack               Stack the environment being activated on top of the
+                                previous active environment, rather replacing the
+                                current active environment with a new one. Currently,
+                                only the PATH environment variable is stacked.
+        """)
+        super(ActivateHelp, self).__init__(message)
+
+
+class DeactivateHelp(Help):
+
+    def __init__(self):
+        message = dals("""
+        usage: conda deactivate [-h]
+
+        Deactivate the current active conda environment.
+
+        Options:
+
+        optional arguments:
+          -h, --help            Show this help message and exit.
+        """)
+        super(DeactivateHelp, self).__init__(message)
+
+
+class GenericHelp(Help):
+
+    def __init__(self, command):
+        message = "help requested for %s" % command
+        super(GenericHelp, self).__init__(message)
+
+
 class CondaSignalInterrupt(CondaError):
     def __init__(self, signum):
         signal_name = get_signal_name(signum)
@@ -174,6 +233,7 @@ class CommandNotFoundError(CondaError):
         activate_commands = {
             'activate',
             'deactivate',
+            'run',
         }
         conda_commands = {
             'clean',
@@ -200,38 +260,29 @@ class CommandNotFoundError(CondaError):
             'render',
             'skeleton',
         }
-        # TODO: Point users to a page at conda-docs, which explains this context in more detail
+        from .base.context import context
+        from .cli.main import init_loggers
+        init_loggers(context)
         if command in activate_commands:
-            from .base.context import context
+            # TODO: Point users to a page at conda-docs, which explains this context in more detail
             builder = ["Your shell has not been properly configured to use 'conda %(command)s'."]
+            if on_win:
+                builder.append(dals("""
+                If using 'conda %(command)s' from a batch script, change your
+                invocation to 'CALL conda.bat %(command)s'.
+                """))
             builder.append(dals("""
-            If your shell is Bash or a Bourne variant, enable conda for the current user with
+            To initialize your shell, run
 
-                $ echo ". %(root_prefix)s/etc/profile.d/conda.sh" >> ~/%(config_file)s
+                $ conda init <SHELL_NAME>
 
-            or, for all users, enable conda with
+            Currently supported shells are:%(supported_shells)s
 
-                $ sudo ln -s %(root_prefix)s/etc/profile.d/conda.sh /etc/profile.d/conda.sh
+            See 'conda init --help' for more information and options.
 
-            The options above will permanently enable the 'conda' command, but they do NOT
-            put conda's base (root) environment on PATH.  To do so, run
-
-                $ conda activate
-
-            in your terminal, or to put the base environment on PATH permanently, run
-
-                $ echo "conda activate" >> ~/%(config_file)s
-
-            Previous to conda 4.4, the recommended way to activate conda was to modify PATH in
-            your ~/%(config_file)s file.  You should manually remove the line that looks like
-
-                export PATH="%(root_prefix)s/bin:$PATH"
-
-            ^^^ The above line should NO LONGER be in your ~/%(config_file)s file! ^^^
+            IMPORTANT: You may need to close and restart your shell after running 'conda init'.
             """) % {
-                'root_prefix': context.root_prefix,
-                'macos_fix': " ''" if sys.platform == 'darwin' else "",
-                'config_file': '.bash_profile' if sys.platform == 'darwin' else '.bashrc',
+                'supported_shells': dashlist(COMPATIBLE_SHELLS),
             })
             message = '\n'.join(builder)
         elif command in build_commands:
@@ -244,9 +295,6 @@ class CommandNotFoundError(CondaError):
             close = get_close_matches(command, choices)
             if close:
                 message += "\nDid you mean 'conda %s'?" % close[0]
-        from .base.context import context
-        from .cli.main import init_loggers
-        init_loggers(context)
         super(CommandNotFoundError, self).__init__(message, command=command)
 
 
@@ -277,6 +325,29 @@ class EnvironmentNameNotFound(CondaError):
         super(EnvironmentNameNotFound, self).__init__(message, environment_name=environment_name)
 
 
+class NoBaseEnvironmentError(CondaError):
+
+    def __init__(self):
+        message = dals("""
+        This conda installation has no default base environment. Use
+        'conda create' to create new environments and 'conda activate' to
+        activate environments.
+        """)
+        super(NoBaseEnvironmentError, self).__init__(message)
+
+
+class DirectoryNotACondaEnvironmentError(CondaError):
+
+    def __init__(self, target_directory):
+        message = dals("""
+        The target directory exists, but it is not a conda environment.
+        Use 'conda create' to convert the directory to a conda environment.
+          target directory: %(target_directory)s
+        """)
+        super(DirectoryNotACondaEnvironmentError, self).__init__(message,
+                                                                 target_directory=target_directory)
+
+
 class CondaEnvironmentError(CondaError, EnvironmentError):
     def __init__(self, message, *args):
         msg = '%s' % message
@@ -285,7 +356,7 @@ class CondaEnvironmentError(CondaError, EnvironmentError):
 
 class DryRunExit(CondaExitZero):
     def __init__(self):
-        msg = 'Dry run exiting'
+        msg = 'Dry run. Exiting.'
         super(DryRunExit, self).__init__(msg)
 
 
@@ -295,11 +366,6 @@ class CondaSystemExit(CondaExitZero, SystemExit):
         super(CondaSystemExit, self).__init__(msg)
 
 
-class SubprocessExit(CondaExitZero):
-    def __init__(self, *args, **kwargs):
-        super(SubprocessExit, self).__init__(*args, **kwargs)
-
-
 class PaddingError(CondaError):
     def __init__(self, dist, placeholder, placeholder_length):
         msg = ("Placeholder of length '%d' too short in package %s.\n"
@@ -313,9 +379,9 @@ class LinkError(CondaError):
 
 
 class CondaOSError(CondaError, OSError):
-    def __init__(self, message):
+    def __init__(self, message, **kwargs):
         msg = '%s' % message
-        super(CondaOSError, self).__init__(msg)
+        super(CondaOSError, self).__init__(msg, **kwargs)
 
 
 class ProxyError(CondaError):
@@ -346,15 +412,55 @@ class CondaKeyError(CondaError, KeyError):
 
 
 class ChannelError(CondaError):
-    def __init__(self, message, *args):
-        msg = '%s' % message
-        super(ChannelError, self).__init__(msg)
+    pass
 
 
 class ChannelNotAllowed(ChannelError):
-    def __init__(self, message, *args):
-        msg = '%s' % message
-        super(ChannelNotAllowed, self).__init__(msg, *args)
+    def __init__(self, channel):
+        from .models.channel import Channel
+        from .common.url import maybe_unquote
+        channel = Channel(channel)
+        channel_name = channel.name
+        channel_url = maybe_unquote(channel.base_url)
+        message = dals("""
+        Channel not included in whitelist:
+          channel name: %(channel_name)s
+          channel url: %(channel_url)s
+        """)
+        super(ChannelNotAllowed, self).__init__(message, channel_url=channel_url,
+                                                channel_name=channel_name)
+
+
+class UnavailableInvalidChannel(ChannelError):
+
+    def __init__(self, channel, error_code):
+        from .models.channel import Channel
+        from .common.url import join_url, maybe_unquote
+        channel = Channel(channel)
+        channel_name = channel.name
+        channel_url = maybe_unquote(channel.base_url)
+        message = dals("""
+        The channel is not accessible or is invalid.
+          channel name: %(channel_name)s
+          channel url: %(channel_url)s
+          error code: %(error_code)d
+
+        You will need to adjust your conda configuration to proceed.
+        Use `conda config --show channels` to view your configuration's current state,
+        and use `conda config --show-sources` to view config file locations.
+        """)
+
+        if channel.scheme == 'file':
+            message += dedent("""
+            As of conda 4.3, a valid channel must contain a `noarch/repodata.json` and
+            associated `noarch/repodata.json.bz2` file, even if `noarch/repodata.json` is
+            empty. Use `conda index %s`, or create `noarch/repodata.json`
+            and associated `noarch/repodata.json.bz2`.
+            """) % join_url(channel.location, channel.name)
+
+        super(UnavailableInvalidChannel, self).__init__(message, channel_url=channel_url,
+                                                        channel_name=channel_name,
+                                                        error_code=error_code)
 
 
 class OperationNotAllowed(CondaError):
@@ -390,6 +496,7 @@ class MD5MismatchError(CondaError):
           expected md5 sum: %(expected_md5sum)s
           actual md5 sum: %(actual_md5sum)s
         """)
+        from .common.url import maybe_unquote
         url = maybe_unquote(url)
         super(MD5MismatchError, self).__init__(message, url=url, target_full_path=target_full_path,
                                                expected_md5sum=expected_md5sum,
@@ -411,6 +518,7 @@ class PackageNotInstalledError(CondaError):
 class CondaHTTPError(CondaError):
     def __init__(self, message, url, status_code, reason, elapsed_time, response=None,
                  caused_by=None):
+        from .common.url import maybe_unquote
         _message = dals("""
         HTTP %(status_code)s %(reason)s for url <%(url)s>
         Elapsed: %(elapsed_time)s
@@ -424,7 +532,7 @@ class CondaHTTPError(CondaError):
         elapsed_time = elapsed_time or '-'
 
         from ._vendor.auxlib.logz import stringify
-        response_details = (stringify(response) or '') if response else ''
+        response_details = (stringify(response, content_max_len=1024) or '') if response else ''
 
         url = maybe_unquote(url)
         if isinstance(elapsed_time, timedelta):
@@ -502,7 +610,6 @@ class UnsatisfiableError(CondaError):
 
     def __init__(self, bad_deps, chains=True):
         from .models.match_spec import MatchSpec
-        from .resolve import dashlist
 
         # Remove any target values from the MatchSpecs, convert to strings
         bad_deps = [list(map(lambda x: str(MatchSpec(x, target=None)), dep)) for dep in bad_deps]
@@ -536,13 +643,13 @@ class UnsatisfiableError(CondaError):
                 chains[key] = ' -> '.join(deps)
             bad_deps = [chains[key] for key in sorted(iterkeys(chains))]
             msg = '''The following specifications were found to be in conflict:%s
-Use "conda info <package>" to see the dependencies for each package.'''
+Use "conda search <package> --info" to see the dependencies for each package.'''
         else:
             bad_deps = [sorted(dep) for dep in bad_deps]
             bad_deps = [', '.join(dep) for dep in sorted(bad_deps)]
             msg = '''The following specifications were found to be incompatible with the
 others, or with the existing package set:%s
-Use "conda info <package>" to see the dependencies for each package.'''
+Use "conda search <package> --info" to see the dependencies for each package.'''
         msg = msg % dashlist(bad_deps)
         super(UnsatisfiableError, self).__init__(msg)
 
@@ -561,13 +668,31 @@ class RemoveError(CondaError):
 
 class DisallowedPackageError(CondaError):
     def __init__(self, package_ref, **kwargs):
-        from .models.records import PackageRef
-        package_ref = PackageRef.from_objects(package_ref)
+        from .models.records import PackageRecord
+        package_ref = PackageRecord.from_objects(package_ref)
         message = ("The package '%(dist_str)s' is disallowed by configuration.\n"
                    "See 'conda config --show disallowed_packages'.")
         super(DisallowedPackageError, self).__init__(message, package_ref=package_ref,
                                                      dist_str=package_ref.dist_str(), **kwargs)
 
+class SpecsConfigurationConflictError(CondaError):
+
+    def __init__(self, requested_specs, pinned_specs, prefix):
+        message = dals("""
+        Requested specs conflict with configured specs.
+          requested specs: {requested_specs_formatted}
+          pinned specs: {pinned_specs_formatted}
+        Use 'conda config --show-sources' to look for 'pinned_specs' and 'track_features'
+        configuration parameters.  Pinned specs may also be defined in the file
+        {pinned_specs_path}.
+        """).format(
+            requested_specs_formatted=dashlist(requested_specs, 4),
+            pinned_specs_formatted=dashlist(pinned_specs, 4),
+            pinned_specs_path=join(prefix, "conda-meta", "pinned"),
+        )
+        super(SpecsConfigurationConflictError, self).__init__(
+            message, requested_specs=requested_specs, pinned_specs=pinned_specs, prefix=prefix,
+        )
 
 class CondaIndexError(CondaError, IndexError):
     def __init__(self, message):
@@ -576,9 +701,9 @@ class CondaIndexError(CondaError, IndexError):
 
 
 class CondaValueError(CondaError, ValueError):
-    def __init__(self, message, *args):
-        msg = '%s' % message
-        super(CondaValueError, self).__init__(msg)
+
+    def __init__(self, message, *args, **kwargs):
+        super(CondaValueError, self).__init__(message, *args, **kwargs)
 
 
 class CondaTypeError(CondaError, TypeError):
@@ -589,9 +714,8 @@ class CondaTypeError(CondaError, TypeError):
 
 class CyclicalDependencyError(CondaError, ValueError):
     def __init__(self, packages_with_cycles, **kwargs):
-        from .resolve import dashlist
-        from .models.records import PackageRef
-        packages_with_cycles = tuple(PackageRef.from_objects(p) for p in packages_with_cycles)
+        from .models.records import PackageRecord
+        packages_with_cycles = tuple(PackageRecord.from_objects(p) for p in packages_with_cycles)
         message = "Cyclic dependencies exist among these items: %s" % dashlist(
             p.dist_str() for p in packages_with_cycles
         )
@@ -686,7 +810,6 @@ class NotWritableError(CondaError, OSError):
 
             In general, it's not advisable to use 'sudo conda'.
             """)
-            import os
             kwargs.update({
                 'uid': os.geteuid(),
                 'gid': os.getegid(),
@@ -694,6 +817,45 @@ class NotWritableError(CondaError, OSError):
         super(NotWritableError, self).__init__(message, **kwargs)
 
 
+class NoWritableEnvsDirError(CondaError):
+
+    def __init__(self, envs_dirs, **kwargs):
+        message = "No writeable envs directories configured.%s" % dashlist(envs_dirs)
+        super(NoWritableEnvsDirError, self).__init__(message, envs_dirs=envs_dirs, **kwargs)
+
+
+class NoWritablePkgsDirError(CondaError):
+
+    def __init__(self, pkgs_dirs, **kwargs):
+        message = "No writeable pkgs directories configured.%s" % dashlist(pkgs_dirs)
+        super(NoWritablePkgsDirError, self).__init__(message, pkgs_dirs=pkgs_dirs, **kwargs)
+
+
+class EnvironmentNotWritableError(CondaError):
+
+    def __init__(self, environment_location, **kwargs):
+        kwargs.update({
+            'environment_location': environment_location,
+        })
+        if on_win:
+            message = dals("""
+            The current user does not have write permissions to the target environment.
+              environment location: %(environment_location)s
+            """)
+        else:
+            message = dals("""
+            The current user does not have write permissions to the target environment.
+              environment location: %(environment_location)s
+              uid: %(uid)s
+              gid: %(gid)s
+            """)
+            kwargs.update({
+                'uid': os.geteuid(),
+                'gid': os.getegid(),
+            })
+        super(EnvironmentNotWritableError, self).__init__(message, **kwargs)
+
+
 class CondaDependencyError(CondaError):
     def __init__(self, message):
         super(CondaDependencyError, self).__init__(message)
@@ -719,10 +881,24 @@ class BinaryPrefixReplacementError(CondaError):
         super(BinaryPrefixReplacementError, self).__init__(message, **kwargs)
 
 
-class InvalidVersionSpecError(CondaError):
-    def __init__(self, invalid_spec):
-        message = "Invalid version spec: %(invalid_spec)s"
-        super(InvalidVersionSpecError, self).__init__(message, invalid_spec=invalid_spec)
+class InvalidSpec(CondaError, ValueError):
+
+    def __init__(self, message, **kwargs):
+        super(InvalidSpec, self).__init__(message, **kwargs)
+
+
+class InvalidVersionSpec(InvalidSpec):
+    def __init__(self, invalid_spec, details):
+        message = "Invalid version '%(invalid_spec)s': %(details)s"
+        super(InvalidVersionSpec, self).__init__(message, invalid_spec=invalid_spec,
+                                                 details=details)
+
+
+class InvalidMatchSpec(InvalidSpec):
+    def __init__(self, invalid_spec, details):
+        message = "Invalid spec '%(invalid_spec)s': %(details)s"
+        super(InvalidMatchSpec, self).__init__(message, invalid_spec=invalid_spec,
+                                               details=details)
 
 
 class EncodingError(CondaError):
@@ -787,17 +963,23 @@ def maybe_raise(error, context):
 
 def print_conda_exception(exc_val, exc_tb=None):
     from .base.context import context
-    if context.debug or context.verbosity > 0:
-        sys.stderr.write(_format_exc(exc_val, exc_tb))
-        sys.stderr.write('\n')
+    rc = getattr(exc_val, 'return_code', None)
+    if (context.debug
+            or context.verbosity > 2
+            or (not isinstance(exc_val, DryRunExit) and context.verbosity > 0)):
+        print(_format_exc(exc_val, exc_tb), file=sys.stderr)
     elif context.json:
-        import json
+        if isinstance(exc_val, DryRunExit):
+            return
         logger = getLogger('conda.stdout' if exc_val.return_code else 'conda.stderr')
         exc_json = json.dumps(exc_val.dump_map(), indent=2, sort_keys=True, cls=EntityEncoder)
         logger.info("%s\n" % exc_json)
     else:
         stderrlog = getLogger('conda.stderr')
-        stderrlog.info("\n%r\n", exc_val)
+        if rc == 0:
+            stderrlog.error("\n%s\n", exc_val)
+        else:
+            stderrlog.error("\n%r\n", exc_val)
 
 
 def _format_exc(exc_val=None, exc_tb=None):
@@ -817,7 +999,7 @@ class ExceptionHandler(object):
     def __call__(self, func, *args, **kwargs):
         try:
             return func(*args, **kwargs)
-        except:
+        except:  # lgtm [py/catch-base-exception]
             _, exc_val, exc_tb = sys.exc_info()
             return self.handle_exception(exc_val, exc_tb)
 
