@@ -1,15 +1,17 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 from errno import EACCES, ELOOP, EPERM
 from io import open
 from logging import getLogger
 import os
-from os import X_OK, access, fstat
 from os.path import basename, dirname, isdir, isfile, join, splitext
 from shutil import copyfileobj, copystat
 import sys
 import tarfile
+import tempfile
 
 from . import mkdir_p
 from .delete import rm_rf
@@ -131,7 +133,7 @@ class ProgressFileWrapper(object):
     def __init__(self, fileobj, progress_update_callback):
         self.progress_file = fileobj
         self.progress_update_callback = progress_update_callback
-        self.progress_file_size = max(1, fstat(fileobj.fileno()).st_size)
+        self.progress_file_size = max(1, os.fstat(fileobj.fileno()).st_size)
         self.progress_max_pos = 0
 
     def __getattr__(self, name):
@@ -205,7 +207,7 @@ def make_menu(prefix, file_path, remove=False):
     try:
         import menuinst
         menuinst.install(join(prefix, win_path_ok(file_path)), remove, prefix)
-    except:
+    except Exception:
         stdoutlog.error("menuinst Exception", exc_info=True)
 
 
@@ -233,7 +235,7 @@ def _is_unix_executable_using_ORIGIN(path):
     if on_win:
         return False
     else:
-        return isfile(path) and not islink(path) and access(path, X_OK)
+        return isfile(path) and not islink(path) and os.access(path, os.X_OK)
 
 
 def _do_softlink(src, dst):
@@ -335,29 +337,53 @@ def create_link(src, dst, link_type=LinkType.hardlink, force=False):
         raise CondaError("Did not expect linktype=%r" % link_type)
 
 
-def compile_pyc(python_exe_full_path, py_full_path, pyc_full_path):
-    if lexists(pyc_full_path):
-        maybe_raise(BasicClobberError(None, pyc_full_path, context), context)
+def compile_multiple_pyc(python_exe_full_path, py_full_paths, pyc_full_paths, prefix, py_ver):
+    py_full_paths = tuple(py_full_paths)
+    pyc_full_paths = tuple(pyc_full_paths)
+    if len(py_full_paths) == 0:
+        return []
 
-    command = '"%s" -Wi -m py_compile "%s"' % (python_exe_full_path, py_full_path)
-    log.trace(command)
-    result = subprocess_call(command, raise_on_error=False)
+    for pyc_full_path in pyc_full_paths:
+        if lexists(pyc_full_path):
+            maybe_raise(BasicClobberError(None, pyc_full_path, context), context)
 
-    if not isfile(pyc_full_path):
-        message = dals("""
-        pyc file failed to compile successfully
-          python_exe_full_path: %s
-          py_full_path: %s
-          pyc_full_path: %s
-          compile rc: %s
-          compile stdout: %s
-          compile stderr: %s
-        """)
-        log.info(message, python_exe_full_path, py_full_path, pyc_full_path,
-                 result.rc, result.stdout, result.stderr)
-        return None
+    fd, filename = tempfile.mkstemp()
+    try:
+        for f in py_full_paths:
+            f = os.path.relpath(f, prefix)
+            if hasattr(f, 'encode'):
+                f = f.encode(sys.getfilesystemencoding(), errors='replace')
+            os.write(fd, f + b"\n")
+        os.close(fd)
+        command = ["-Wi", "-m", "compileall", "-q", "-l", "-i", filename]
+        # if the python version in the prefix is 3.5+, we have some extra args.
+        #    -j 0 will do the compilation in parallel, with os.cpu_count() cores
+        if int(py_ver[0]) >= 3 and int(py_ver.split('.')[1]) > 5:
+            command.extend(["-j", "0"])
+        command = '"%s" ' % python_exe_full_path + " ".join(command)
+        log.trace(command)
+        result = subprocess_call(command, raise_on_error=False, path=prefix)
+    finally:
+        os.remove(filename)
 
-    return pyc_full_path
+    created_pyc_paths = []
+    for py_full_path, pyc_full_path in zip(py_full_paths, pyc_full_paths):
+        if not isfile(pyc_full_path):
+            message = dals("""
+            pyc file failed to compile successfully
+            python_exe_full_path: %s
+            py_full_path: %s
+            pyc_full_path: %s
+            compile rc: %s
+            compile stdout: %s
+            compile stderr: %s
+            """)
+            log.info(message, python_exe_full_path, py_full_path, pyc_full_path,
+                     result.rc, result.stdout, result.stderr)
+        else:
+            created_pyc_paths.append(pyc_full_path)
+
+    return created_pyc_paths
 
 
 def create_package_cache_directory(pkgs_dir):
@@ -374,3 +400,23 @@ def create_package_cache_directory(pkgs_dir):
         else:
             raise
     return True
+
+
+def create_envs_directory(envs_dir):
+    # returns False if envs directory cannot be created
+
+    # The magic file being used here could change in the future.  Don't write programs
+    # outside this code base that rely on the presence of this file.
+    # This value is duplicated in conda.base.context._first_writable_envs_dir().
+    envs_dir_magic_file = join(envs_dir, '.conda_envs_dir_test')
+    try:
+        log.trace("creating envs directory '%s'", envs_dir)
+        sudo_safe = expand(envs_dir).startswith(expand('~'))
+        touch(join(envs_dir, envs_dir_magic_file), mkdir=True, sudo_safe=sudo_safe)
+    except (IOError, OSError) as e:
+        if e.errno in (EACCES, EPERM):
+            log.trace("cannot create envs directory '%s'", envs_dir)
+            return False
+        else:
+            raise
+    return True
