@@ -1,3 +1,6 @@
+# -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 from ast import literal_eval
@@ -14,22 +17,17 @@ import warnings
 
 from . import __version__ as CONDA_VERSION
 from ._vendor.auxlib.ish import dals
+from ._vendor.toolz import groupby, take
 from .base.constants import DEFAULTS_CHANNEL_NAME
 from .base.context import context
 from .common.compat import ensure_text_type, iteritems, open, text_type
 from .common.path import paths_equal
-from .core.prefix_data import PrefixData, linked
+from .core.prefix_data import PrefixData
 from .exceptions import CondaHistoryError, CondaUpgradeError, NotWritableError
 from .gateways.disk.update import touch
-from .models.dist import Dist
+from .models.dist import dist_str_to_quad
 from .models.version import VersionOrder, version_relation_re
-from .resolve import MatchSpec
-
-try:
-    from cytoolz.itertoolz import groupby, take
-except ImportError:  # pragma: no cover
-    from ._vendor.toolz.itertoolz import groupby, take  # NOQA
-
+from .models.match_spec import MatchSpec
 
 log = logging.getLogger(__name__)
 
@@ -53,8 +51,7 @@ def pretty_diff(diff):
     removed = {}
     for s in diff:
         fn = s[1:]
-        dist = Dist(fn)
-        name, version, _, channel = dist.quad
+        name, version, _, channel = dist_str_to_quad(fn)
         if channel != DEFAULTS_CHANNEL_NAME:
             version += ' (%s)' % channel
         if s.startswith('-'):
@@ -112,7 +109,8 @@ class History(object):
                 warnings.warn("Error in %s: %s" % (self.path, e),
                               CondaHistoryWarning)
                 return
-            curr = set(map(str, linked(self.prefix)))
+            pd = PrefixData(self.prefix)
+            curr = set(prefix_rec.dist_str() for prefix_rec in pd.iter_records())
             self.write_changes(last, curr)
         except EnvironmentError as e:
             if e.errno in (EACCES, EPERM):
@@ -233,7 +231,7 @@ class History(object):
 
         conda_versions_from_history = tuple(x['conda_version'] for x in res
                                             if 'conda_version' in x)
-        if conda_versions_from_history:
+        if conda_versions_from_history and not context.allow_conda_downgrades:
             minimum_conda_version = sorted(conda_versions_from_history, key=VersionOrder)[-1]
             minimum_major_minor = '.'.join(take(2, minimum_conda_version.split('.')))
             current_major_minor = '.'.join(take(2, CONDA_VERSION.split('.')))
@@ -303,6 +301,8 @@ class History(object):
         return the state, i.e. the set of distributions, for a given revision,
         defaults to latest (which is the same as the current state when
         the log file is up-to-date)
+
+        Returns a list of dist_strs
         """
         states = self.construct_states()
         if not states:
@@ -334,7 +334,7 @@ class History(object):
             removed = {}
             if is_diff(content):
                 for pkg in content:
-                    name, version, build, channel = Dist(pkg[1:]).quad
+                    name, version, build, channel = dist_str_to_quad(pkg[1:])
                     if pkg.startswith('+'):
                         added[name.lower()] = (version, build, channel)
                     elif pkg.startswith('-'):
@@ -391,4 +391,4 @@ if __name__ == '__main__':
     # Don't use in context manager mode---it augments the history every time
     h = History(sys.prefix)
     pprint(h.get_user_requests())
-    print(h.get_requested_specs())
+    print(h.get_requested_specs_map())
