@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+# Copyright (C) 2012 Anaconda, Inc
+# SPDX-License-Identifier: BSD-3-Clause
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 from copy import copy
@@ -6,6 +8,7 @@ from itertools import chain
 from logging import getLogger
 
 from .._vendor.boltons.setutils import IndexedSet
+from .._vendor.toolz import concat, concatv, drop
 from ..base.constants import DEFAULTS_CHANNEL_NAME, MAX_CHANNEL_PRIORITY, UNKNOWN_CHANNEL
 from ..base.context import context
 from ..common.compat import ensure_text_type, isiterable, iteritems, odict, with_metaclass
@@ -14,13 +17,6 @@ from ..common.url import (Url, has_scheme, is_url, join_url, path_to_url,
                           split_conda_url_easy_parts, split_platform, split_scheme_auth_token,
                           urlparse)
 
-try:
-    from cytoolz.functoolz import excepts
-    from cytoolz.itertoolz import concat, concatv, drop
-except ImportError:  # pragma: no cover
-    from .._vendor.toolz.functoolz import excepts  # NOQA
-    from .._vendor.toolz.itertoolz import concat, concatv, drop  # NOQA
-
 log = getLogger(__name__)
 
 
@@ -133,7 +129,8 @@ class Channel(object):
                 location, name = ca.location, test_url.replace(ca.location, '', 1)
             else:
                 url_parts = urlparse(test_url)
-                location, name = Url(host=url_parts.host, port=url_parts.port).url, url_parts.path
+                location = Url(host=url_parts.host, port=url_parts.port).url
+                name = url_parts.path or ''
             return Channel(scheme=scheme, auth=auth, location=location, token=token,
                            name=name.strip('/'))
         else:
@@ -142,25 +139,38 @@ class Channel(object):
 
     @property
     def canonical_name(self):
+        try:
+            return self.__canonical_name
+        except AttributeError:
+            pass
+
         for multiname, channels in iteritems(context.custom_multichannels):
             for channel in channels:
                 if self.name == channel.name:
-                    return multiname
+                    cn = self.__canonical_name = multiname
+                    return cn
 
         for that_name in context.custom_channels:
             if self.name and tokenized_startswith(self.name.split('/'), that_name.split('/')):
-                return self.name
+                cn = self.__canonical_name = self.name
+                return cn
 
-        if any(c.location == self.location
-               for c in concatv((context.channel_alias,), context.migrated_channel_aliases)):
-            return self.name
+        if any(c.location == self.location for c in concatv(
+                (context.channel_alias,),
+                context.migrated_channel_aliases,
+        )):
+            cn = self.__canonical_name = self.name
+            return cn
 
         # fall back to the equivalent of self.base_url
         # re-defining here because base_url for MultiChannel is None
         if self.scheme:
-            return "%s://%s" % (self.scheme, join_url(self.location, self.name))
+            cn = self.__canonical_name = "%s://%s" % (self.scheme,
+                                                      join_url(self.location, self.name))
+            return cn
         else:
-            return join_url(self.location, self.name).lstrip('/')
+            cn = self.__canonical_name = join_url(self.location, self.name).lstrip('/')
+            return cn
 
     def urls(self, with_credentials=False, subdirs=None):
         if subdirs is None:
@@ -241,14 +251,7 @@ class Channel(object):
             return base
 
     def __repr__(self):
-        return ("Channel(scheme=%r, auth=%r, location=%r, token=%r, name=%r, platform=%r, "
-                "package_filename=%r)" % (self.scheme,
-                                          self.auth and "%s:<PASSWORD>" % self.auth.split(':')[0],
-                                          self.location,
-                                          self.token and "<TOKEN>",
-                                          self.name,
-                                          self.platform,
-                                          self.package_filename))
+        return 'Channel("%s")' % (join_url(self.name, self.subdir) if self.subdir else self.name)
 
     def __eq__(self, other):
         if isinstance(other, Channel):
